package bagaturchess.engines.base.cfg;


import java.util.ArrayList;
import java.util.List;

import bagaturchess.bitboard.api.IBoardConfig;
import bagaturchess.bitboard.impl.utils.ReflectionUtils;
import bagaturchess.search.api.IEvalConfig;
import bagaturchess.search.api.IRootSearchConfig;
import bagaturchess.search.api.ISearchConfig_AB;
import bagaturchess.uci.api.IUCIOptionsProvider;
import bagaturchess.uci.api.IUCIOptionsRegistry;
import bagaturchess.uci.impl.commands.options.UCIOption;
import bagaturchess.uci.impl.commands.options.UCIOptionSpin;


public class RootSearchConfig_BaseImpl implements IRootSearchConfig, IUCIOptionsProvider {

	
	private UCIOption[] options = new UCIOption[] {
			new UCIOptionSpin("MultiPV", 1.0, "type spin default 1 min 1 max 100", 1),
	};
	
	
	private String searchImpl_ClassName;
	private ISearchConfig_AB searchImpl_ConfigObj;
	
	private IBoardConfig boardCfg;
	private IEvalConfig evalCfg;
	
	private int multiPVsCount = 1;
	
	
	public RootSearchConfig_BaseImpl(String[] args) {
		searchImpl_ClassName = args[0];
		
		if (args.length <= 4) {
			searchImpl_ConfigObj = (ISearchConfig_AB) ReflectionUtils.createObjectByClassName_NoArgsConstructor(args[1]);
		}
		
		
		if (args.length > 2) {
			boardCfg = (IBoardConfig) ReflectionUtils.createObjectByClassName_NoArgsConstructor(args[2]);
			
			if (args.length > 3) {
				evalCfg = (IEvalConfig) ReflectionUtils.createObjectByClassName_NoArgsConstructor(args[3]);
				
				if (args.length > 4) {
					//Extract search params
					List<String> params = new ArrayList<String>();
					for (int i=4; i<args.length; i++) {
						if (args[i] != null && args[i].contains("=")) {
							params.add(args[i]);
						}
					}
					searchImpl_ConfigObj = (ISearchConfig_AB) ReflectionUtils.createObjectByClassName_StringsConstructor(args[1], params.toArray(new String[0]));
				}
			}
		}
	}
	
	
	@Override
	public String getSearchClassName() {
		return searchImpl_ClassName;
	}
	
	
	@Override
	public ISearchConfig_AB getSearchConfig() {
		return searchImpl_ConfigObj;
	}
	
	
	@Override
	public IBoardConfig getBoardConfig() {
		return boardCfg;
	}
	
	
	@Override
	public IEvalConfig getEvalConfig() {
		return evalCfg;
	}
	
	
	@Override
	public double getTPTUsagePercent() {
		return 0.35;//0.45;
	}
	
	
	@Override
	public double getEvalCacheUsagePercent() {
		return 0.35;//0.35;
	}
	
	
	@Override
	public double getPawnsCacheUsagePercent() {
		return 0.3;//0.2;
	}
	
	
	@Override
	public int getMultiPVsCount() {
		return multiPVsCount;
	}
	
	
	@Override
	public void registerProviders(IUCIOptionsRegistry registry) {
		
		registry.registerProvider(this);
		
		if (searchImpl_ConfigObj instanceof IUCIOptionsProvider) {
			registry.registerProvider((IUCIOptionsProvider) searchImpl_ConfigObj);
		}
		
		if (evalCfg instanceof IUCIOptionsProvider) {
			registry.registerProvider((IUCIOptionsProvider) evalCfg);
		}
	}
	
	
	@Override
	public UCIOption[] getSupportedOptions() {
		return options;
	}
	
	
	@Override
	public boolean applyOption(UCIOption option) {
		if ("MultiPV".equals(option.getName())) {
			multiPVsCount = (int) ((Double) option.getValue()).doubleValue();
			return true;
		}
		
		return false;
	}
}
